from bfieldtools import contour

import pytest

from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
    assert_equal,
)


def test_scalar_contour():

    assert True


def test_simplify_contour():

    assert True
