.. raw:: html

    <div id="index_carousel_tn" class="carousel-outer">
      <div id="examples_carousel" class="carousel slide" data-ride="carousel">
        <ol id="mne_index_indicators" class="carousel-indicators">
          <li data-target="#examples_carousel" data-slide-to="0" class="active"></li>
          <li data-target="#examples_carousel" data-slide-to="1" class=""></li>
          <li data-target="#examples_carousel" data-slide-to="2" class=""></li>
          <li data-target="#examples_carousel" data-slide-to="3" class=""></li>
        </ol>
        <!-- Carousel items -->
        <div class="carousel-inner">

          <div class="item active">
            <div class="chopper container" style="background-image: url(_images/sphx_glr_biplanar_coil_design_002.png)" title="Coil design">
              <div class="carousel-caption">
                <h3>Coil design</h3>
                <p>Stream-function-based coil design in arbitrary geometries and with multiple constraints
                   <a class="btn btn-default btn-xs" href="auto_examples/index.html#coil-design" role="button">Check it out<i class="fa fa-arrow-right"></i></a></p>
              </div>
            </div>
          </div>
          <div class="item">
            <div class="chopper container" style="background-image: url(_images/sphx_glr_representation_comparison_002.png)" title="Clusters">
              <div class="carousel-caption">
                <h3>Surface harmonics</h3>
                <p>Surface harmonics basis for compact representation of surface currents
                   <a class="btn btn-default btn-xs" href="auto_examples/publication_physics/representation_comparison.html" role="button">Check it out<i class="fa fa-arrow-right"></i></a></p>
              </div>
            </div>
          </div>
          <div class="item">
            <div class="chopper container" style="background-image: url(_images/sphx_glr_pub_magnetically_shielded_biplanar_coil_design_005.png)" title="Clusters">
              <div class="carousel-caption">
                <h3>Modeling of magnetic shielding</h3>
                <p>Arbitrary closed geometries, modeling and prospective compensation
                   <a class="btn btn-default btn-xs" href="auto_examples/coil_design/head_gradient_coil_design.html" role="button">Check it out<i class="fa fa-arrow-right"></i></a></p>
              </div>
            </div>
          </div>
          <div class="item">
            <div class="chopper container" style="background-image: url(_images/sphx_glr_thermal_noise_simulation_003.png)" title="Clusters">
              <div class="carousel-caption">
                <h3>Thermal magnetic noise</h3>
                <p>Computation of thermal magnetic noise, including frequency-dependence 
                   <a class="btn btn-default btn-xs" href="auto_examples/index.html#thermal-noise-computation" role="button">Check it out<i class="fa fa-arrow-right"></i></a></p>
              </div>
            </div>
          </div>
        </div>
        <!-- Carousel nav -->
        <a class="carousel-control left" href="#examples_carousel" role="button" data-slide="prev"><span class="glyphicon glyphicon-chevron-left"></span></a>
        <a class="carousel-control right" href="#examples_carousel" role="button" data-slide="next"><span class="glyphicon glyphicon-chevron-right"></span></a>
      </div>
    </div>
